@@ -4,31 +4,34 @@
 #' specified forecast horizons and (b) forecast into the future with a trained ML model.
 #'
 #' @param data A data.frame with the (a) target to be forecasted and (b) features/predictors. An optional date column can be given in the
-#' \code{dates} argument (required for grouped time-series). Note that forecastML only works with regularly spaced time/date intervals and that missing
+#' \code{dates} argument (required for grouped time series). Note that forecastML only works with regularly spaced date/time intervals and that missing
 #' rows--usually due to periods when no data was collected--will result in poorly trained models due to incorrect feature lags.
 #' Use \code{\link{fill_gaps}} to fill in any missing rows/data prior to running this function.
 #' @param type The type of dataset to return--(a) model training or (b) forecast prediction. The default is \code{train}.
 #' @param outcome_col The column index--an integer--of the target to be forecasted. Forecasting only one outcome column is allowed at present, however,
-#' groups of time-series can be forecasted if they are stacked vertically in a long dataset and the \code{groups}, \code{dates},
+#' groups of time series can be forecasted if they are stacked vertically in a long dataset and the \code{groups}, \code{dates},
 #' and \code{frequency} arguments are specified.
 #' @param horizons A numeric vector of one or more forecast horizons, h, measured in input dataset rows. For each horizon, 1:h
 #' forecasts are returned (e.g., \code{horizons = 12} trains a model to minimize 1 to 12-step-ahead error and returns forecasts
 #' for 1:12 steps into the future). If \code{dates} are given, a horizon of 1, for example, would equal 1 * \code{frequency} in calendar time.
 #' @param lookback A numeric vector giving the lags--in dataset rows--for creating the lagged features. All non-grouping,
-#' non-static, and non-dynamic features in the input dataset, \code{data}, are lagged by the same values. Lags that don't
-#' support direct forecasting for a given horizon are silently dropped. Either \code{lookback} or \code{lookback_control} need to be specified.
+#' non-static, and non-dynamic features in the input dataset, \code{data}, are lagged by the same values.
+#' For lagged features, lag values that don't support direct forecasting for a given forecast horizon
+#' (e.g, a lookback of 3 for a 4-step-ahead horizon) are silently dropped from the data.frame.
+#' Either \code{lookback} or \code{lookback_control} need to be specified.
 #' @param lookback_control A list of numeric vectors, specifying potentially unique lags for each feature. The length
-#' of the list should equal \code{ncol(data)} and be ordered the same as the columns in \code{data}. For grouped time-series, lags for the grouping column(s)
-#' and static feature columns should have a \code{lookback_control} value of 0. \code{list(NULL)} \code{lookback_control} values drop columns
-#' from the input dataset. Lags that don't support direct forecasting for a given horizons
-#' are silently dropped. Either \code{lookback} or \code{lookback_control} need to be specified.
-#' @param dates A vector or 1-column data.frame of dates with class 'Date'. The length of dates should equal \code{nrow(data)}. Required if \code{groups}
-#' are given.
-#' @param frequency Date frequency. A string taking the same input as \code{base::seq.Date(..., by = "frequency")} e.g., '1 month', '7 days', '10 years' etc.
-#' The highest frequency supported at present is '1 day'. Required if \code{dates} are given.
+#' of the list should equal \code{ncol(data)} and be ordered the same as the columns in \code{data}. Lag values for any grouping,
+#' static, or dynamic feature columns are automatically coerced to 0 nd not lagged. \code{list(NULL)} \code{lookback_control} values drop columns
+#' from the input dataset. For lagged features, lag values that don't support direct forecasting for a given forecast horizon
+#' are silently dropped from the data.frame. Either \code{lookback} or \code{lookback_control} need to be specified.
+#' @param dates A vector or 1-column data.frame of dates/times with class 'Date' or 'POSIXt'. The length
+#' of 'dates' should equal \code{nrow(data)}. Required if \code{groups} are given.
+#' @param frequency Date/time frequency. Required if \code{dates} are given. A string taking the same input as \code{base::seq.Date(..., by = "frequency")} or
+#' \code{base::seq.POSIXt(..., by = "frequency")} e.g., '1 hour', '1 month', '7 days', '10 years' etc.
+#' The highest frequency supported at present is '1 sec'.
 #' @param dynamic_features A character vector of column names that identify features that change through time but which are not lagged (e.g., weekday or year).
 #' If \code{type = "forecast"}, these features will receive \code{NA} values; though, they can be filled in by the user after running this function.
-#' @param groups A character vector of column names that identify the groups/hierarchies when multiple time-series are present. These columns are used as model features but
+#' @param groups A character vector of column names that identify the groups/hierarchies when multiple time series are present. These columns are used as model features but
 #' are not lagged. Note that combining feature lags with grouped time series will result in \code{NA} values throughout the data.
 #' @param static_features For grouped time series only. A character vector of column names that identify features that do not change through time.
 #' These columns are not lagged. If \code{type = "forecast"}, these features will be filled forward using the most recent value for the group.
@@ -36,6 +39,8 @@
 #' \code{multisession} or \code{multicore} futures are especially useful for (a) grouped time series with many groups and
 #' (b) high-dimensional datasets with many lags per feature. Run \code{future::plan(future::multiprocess)} prior to this
 #' function to set up multissession or multicore parallel dataset creation.
+#' @param keep_rows Boolean. For non-grouped time series, keep the \code{1:max(lookback)} rows at the beginning of the time series. These rows will
+#' contain missing values for lagged features that "look back" before the start of the dataset.
 #' @return An S3 object of class 'lagged_df' or 'grouped_lagged_df': A list of data.frames with new columns for the lagged/non-lagged features.
 #' The length of the returned list is equal to the number of forecast horizons and is in the order of
 #' horizons supplied to the \code{horizons} argument. Horizon-specific datasets can be accessed with
@@ -44,7 +49,7 @@
 #' The contents of the returned data.frames are as follows:
 #'
 #' \describe{
-#'   \item{\strong{type = 'train', non-grouped:}}{A data.frame with the outcome and lagged features with the first \code{1:max(lookback)} rows removed.}
+#'   \item{\strong{type = 'train', non-grouped:}}{A data.frame with the outcome and lagged/dynamic features.}
 #'   \item{\strong{type = 'train', grouped:}}{A data.frame with the outcome and unlagged grouping columns followed by lagged, dynamic, and static features.}
 #'   \item{\strong{type = 'forecast', non-grouped:}}{(1) An 'index' column giving the row index or date of the
 #'   forecast periods (e.g., a 100 row non-date-based training dataset would start with an index of 101). (2) A 'horizon' column
@@ -52,7 +57,7 @@
 #'   'train', non-grouped dataset.}
 #'   \item{\strong{type = 'forecast', grouped:}}{(1) An 'index' column giving the date of the
 #'   forecast periods. The first forecast date for each group is the maximum date from the \code{dates} argument
-#'   + 1 * \code{frequency} which is the user-supplied date frequency.(2) A 'horizon' column that indicates
+#'   + 1 * \code{frequency} which is the user-supplied date/time frequency.(2) A 'horizon' column that indicates
 #'   the forecast period from \code{1:max(horizons)}. (3) Lagged, static, and dynamic features identical to the 'train', grouped dataset.}
 #' }
 #' @section Attributes:
@@ -91,17 +96,19 @@
 #' }
 #' @example /R/examples/example_create_lagged_df.R
 #'
+#' @import dplyr
 #' @import ggplot2
 #' @importFrom stats as.formula complete.cases cov sd
 #' @importFrom magrittr %>%
 #' @importFrom lubridate %m-%
 #' @importFrom rlang .data
 #' @importFrom purrr map2
+#' @importFrom data.table :=
 #'
 #' @export
 create_lagged_df <- function(data, type = c("train", "forecast"), outcome_col = 1L, horizons, lookback = NULL,
                              lookback_control = NULL, dates = NULL, frequency = NULL, dynamic_features = NULL,
-                             groups = NULL, static_features = NULL, use_future = FALSE) {
+                             groups = NULL, static_features = NULL, use_future = FALSE, keep_rows = FALSE) {
 
   if (!methods::is(data, c("data.frame"))) {
     stop("The 'data' argument takes an object of class 'data.frame'.")
@@ -120,7 +127,7 @@ create_lagged_df <- function(data, type = c("train", "forecast"), outcome_col =
   }
 
   if (!all(horizons < nrow(data))) {
-    stop("The forecast horizons needs to be less than nrow(data).")
+    stop("The forecast horizons need to be less than nrow(data).")
   }
 
   if (all(is.null(lookback), is.null(lookback_control))) {
@@ -139,7 +146,7 @@ create_lagged_df <- function(data, type = c("train", "forecast"), outcome_col =
     stop("The highest lookback needs to be >= the shortest forecast horizons to allow for direct forecasting with lagged features.")
   }
 
-  if (length(horizons) == 1 && !is.null(lookback_control)) {  # A 1-horizons, non-nested lookback_control of feature lags.
+  if (length(horizons) == 1 && !is.null(lookback_control)) {  # A 1-horizon, non-nested lookback_control of feature lags.
 
     # Check if there is one list location for each feature in the dataset.
     if (length(lookback_control) != ncol(data)) {
@@ -168,18 +175,23 @@ create_lagged_df <- function(data, type = c("train", "forecast"), outcome_col =
     dates
   }
 
-  if (!is.null(dates) && !methods::is(dates, "Date") && length(dates) != nrow(data)) {
-    stop("The 'dates' argument needs to be a vector or 1 column data.frame of length nrow(data)
-         of dates with class `Date`.")
+  if (!is.null(dates) && !methods::is(dates, "Date") && !methods::is(dates, c("POSIXt"))) {
+    stop("The 'dates' argument should be an object of class 'Date' or 'POSIXt'.")
+  }
+
+  if (!is.null(dates) && length(dates) != nrow(data)) {
+    stop("The 'dates' argument needs to be a vector or 1-column data.frame of length nrow(data)
+         of dates with class 'Date' or 'POSIXt'.")
   }
 
   if (!is.null(dates) && is.null(frequency)) {
-    stop("The 'frequency' argument needs to be specified along with the `dates` argument.
-         It takes the same input as `base::seq.Date()` e.g., '1 month', '7 days', etc.")
+    stop("The 'frequency' argument needs to be specified along with the 'dates' argument.
+         See base::seq.Date() or base::seq.POSIXt() for valid date/time frequencies e.g., '1 hour',
+         '1 day', '3 months', '10 years' etc.")
   }
 
-  if (!is.null(frequency) && !grepl("day|week|month|quarter|year", frequency)) {
-    stop("The 'frequency' argument should be a string containing one of 'day', 'week',
+  if (!is.null(frequency) && !grepl("sec|min|hour|day|week|month|quarter|year", frequency)) {
+    stop("The 'frequency' argument should be a string containing one of 'sec', 'min', 'hour', 'day', 'week',
          'month', 'quarter', or 'year'. This can optionally be preceded by a positive integer and a space
          and/or followed by an 's'.")
   }
@@ -280,13 +292,12 @@ create_lagged_df <- function(data, type = c("train", "forecast"), outcome_col =
     })  # Impossible lags for lagged features have been removed.
   }  # Impossible lags in 'lookback_control' have been removed.
   #--------------------------------------------------------------------------
-
   # This will be used to remove the rows with NAs in our new lagged predictor dataset--rows 1:lookback_max at the begnning of the dataset.
   # This is only used for non-grouped datasets and allows easy forecasting with methods that can't handle NA values.
   if (!is.null(lookback)) {
     lookback_max <- max(lookback, na.rm = TRUE)
   } else {
-    # TO-do: Greater flexibility could be added in the case of different horizons having different max lags. At present, the lags for longer forecast
+    # TO-DO: Greater flexibility could be added in the case of different horizons having different max lags. At present, the lags for longer forecast
     # horizons are also used to remove dataset rows from input datasets with shorter forecast horizons, which may only use shorter lags. This amounts
     # to unnecessarily discarding training data for short-term forecast horizons when these models are trained alongside long-term models.
     lookback_max <- max(unlist(lookback_control), na.rm = TRUE)
@@ -317,7 +328,7 @@ create_lagged_df <- function(data, type = c("train", "forecast"), outcome_col =
         lookback_over_horizon <- lookback[lookback >= forecast_horizon]
       }
 
-      data_x <- lapply_function(1:ncol(data), function(j) {
+      data_x <- lapply_function(1:ncol(data), function(j, future.packages, future.seed) {
         # Only create lagged features that allow direct forecasting to max(i)--unique lags for each feature.
         if (!is.null(lookback_control)) {
           # As a convenience to the user, a single-horizons forecast that uses a custom lookback doesn't need to be a nested list.
@@ -370,9 +381,12 @@ create_lagged_df <- function(data, type = c("train", "forecast"), outcome_col =
 
               } else {  # This feature is not a grouping/dynamic/static feature and we'll compute lagged versions.
 
-                data_x <- data[, c(groups, var_names[j]), drop = FALSE] %>%
+                data_dt <- dtplyr::lazy_dt(data[, c(groups, var_names[j]), drop = FALSE])
+
+                data_x <- data_dt %>%
                   dplyr::group_by_at(dplyr::vars(groups)) %>%
-                  dplyr::mutate_at(dplyr::vars(var_names[j]), lag_functions)
+                  dplyr::mutate_at(dplyr::vars(var_names[j]), lag_functions) %>%
+                  dplyr::as_tibble()
                 data_x <- data_x[, (ncol(data_x) - length(lag_functions) + 1):ncol(data_x), drop = FALSE]  # Keep only the lagged feature columns.
               }
 
@@ -385,8 +399,11 @@ create_lagged_df <- function(data, type = c("train", "forecast"), outcome_col =
 
               } else {
 
-                data_x <- data[, var_names[j], drop = FALSE] %>%
-                  dplyr::mutate_at(dplyr::vars(var_names[j]), .funs = lag_functions)
+                data_dt <- dtplyr::lazy_dt(data[, var_names[j], drop = FALSE])
+
+                data_x <- data_dt %>%
+                  dplyr::mutate_at(dplyr::vars(var_names[j]), lag_functions) %>%
+                  dplyr::as_tibble()
                 data_x <- data_x[, (ncol(data_x) - length(lag_functions) + 1):ncol(data_x), drop = FALSE]  # Keep only the lagged feature columns.
               }
             }  # End feature-level lag creation across `lookback_over_horizon`.
@@ -408,7 +425,7 @@ create_lagged_df <- function(data, type = c("train", "forecast"), outcome_col =
         }
 
         data_x
-      })  # End loop 'j', the creation of lagged features for a given forecast model horizons.
+      }, future.packages = "dplyr", future.seed = 1)  # End loop 'j', the creation of lagged features for a given forecast model horizons.
 
       data_x <- dplyr::bind_cols(data_x)  # A single data.frame of lags for all features at a given forecast model horizons.
 
@@ -423,7 +440,10 @@ create_lagged_df <- function(data, type = c("train", "forecast"), outcome_col =
       # If the forecast is grouped, leave the NAs in the dataset for the user because the ML model used for these
       # cases will likely handle NA values.
       if (is.null(groups)) {
-        data_out <- data_out[-(1:lookback_max), ]  # To-do: make this more flexible for multiple forecast model horizons.
+
+        if (isFALSE(keep_rows)) {
+          data_out <- data_out[-(1:lookback_max), ]  # Remove the first rows with NAs in lagged features.
+        }
       }
 
       attr(data_out, "horizons") <- forecast_horizon
@@ -452,8 +472,7 @@ create_lagged_df <- function(data, type = c("train", "forecast"), outcome_col =
         lookback_over_horizon <- lookback[lookback >= forecast_horizon]
       }
 
-      data_x <- lapply_function(1:ncol(data), function(j) {
-
+      data_x <- lapply_function(1:ncol(data), function(j, future.packages, future.seed) {
         # Only create lagged features that allow direct forecasting to max(i)--unique lags for each feature.
         if (!is.null(lookback_control)) {
           # As a convenience to the user a single-horizons forecast that uses a custom lookback doesn't need to be a nested list.
@@ -511,13 +530,17 @@ create_lagged_df <- function(data, type = c("train", "forecast"), outcome_col =
               # Now that we know the column is a grouping column, is it the first grouping column?
               if (j == min(group_cols)) {
 
+                #data_dt <- dtplyr::lazy_dt(data[, c(groups, static_features), drop = FALSE])
+                #data_dt <- data[, c(groups, static_features), drop = FALSE]
+
                 data_x <- data[, c(groups, static_features), drop = FALSE] %>%
                   dplyr::group_by_at(dplyr::vars(groups)) %>%
                   dplyr::mutate("row_number" = 1:dplyr::n(),
                                 "max_row_number" = max(.data$row_number, na.rm = TRUE)) %>%
                   dplyr::filter(.data$row_number == .data$max_row_number) %>%
                   dplyr::ungroup() %>%
-                  dplyr::select(!!groups, !!static_features, .data$max_row_number)
+                  dplyr::select(!!groups, !!static_features, .data$max_row_number) %>%
+                  dplyr::as_tibble()
 
                 # Create the same, static dataset for forecasting into the future, the only difference
                 # being an index which indicates the forecast horizons.
@@ -545,6 +568,9 @@ create_lagged_df <- function(data, type = c("train", "forecast"), outcome_col =
 
               if (!var_names[j] %in% c(dynamic_features)) {  # Lagged, non-dynamic features
 
+                #data_dt <- dtplyr::lazy_dt(data_x)
+                #data_dt <- data_x
+
                 data_x <- data_x %>%
                   dplyr::group_by_at(dplyr::vars(groups)) %>%
                   dplyr::mutate("row_number" = 1:dplyr::n()) %>%
@@ -555,10 +581,14 @@ create_lagged_df <- function(data, type = c("train", "forecast"), outcome_col =
                   dplyr::mutate("horizon" = rev(.data$horizon),
                                 "row_number" = .data$max_row_number + .data$horizon) %>%
                   dplyr::ungroup() %>%
-                  dplyr::select(.data$row_number, .data$horizon, groups, names(lag_functions))
+                  dplyr::select(.data$row_number, .data$horizon, groups, names(lag_functions)) %>%
+                  dplyr::as_tibble()
 
               } else {  # Assign NA values to all future dynamic features. There's some unnecessary computation here.
 
+                #data_dt <- dtplyr::lazy_dt(data_x)
+                #data_dt <- data_x
+
                 data_x <- data_x %>%
                   dplyr::group_by_at(dplyr::vars(groups)) %>%
                   dplyr::mutate("row_number" = 1:dplyr::n()) %>%
@@ -568,7 +598,8 @@ create_lagged_df <- function(data, type = c("train", "forecast"), outcome_col =
                   dplyr::mutate("horizon" = rev(.data$horizon),
                                 "row_number" = .data$max_row_number + .data$horizon) %>%
                   dplyr::ungroup() %>%
-                  dplyr::select(.data$row_number, .data$horizon, groups)
+                  dplyr::select(.data$row_number, .data$horizon, groups) %>%
+                  dplyr::as_tibble()
 
                 data_x[[var_names[j]]] <- NA
               }
@@ -583,6 +614,9 @@ create_lagged_df <- function(data, type = c("train", "forecast"), outcome_col =
 
             if (!var_names[j] %in% c(dynamic_features)) {  # Lagged, non-dynamic features.
 
+              #data_dt <- dtplyr::lazy_dt(data[, var_names[j], drop = FALSE])
+              #data_dt <- data[, var_names[j], drop = FALSE]
+
               data_x <- data[, var_names[j], drop = FALSE] %>%
                 dplyr::mutate("row_number" = 1:dplyr::n()) %>%
                 dplyr::mutate_at(dplyr::vars(var_names[j]), lag_functions) %>%
@@ -591,10 +625,14 @@ create_lagged_df <- function(data, type = c("train", "forecast"), outcome_col =
                 dplyr::filter(.data$horizon <= forecast_horizon) %>%
                 dplyr::mutate("horizon" = rev(.data$horizon),
                               "row_number" = .data$max_row_number + .data$horizon) %>%
-                dplyr::select(.data$row_number, .data$horizon, groups, names(lag_functions))
+                dplyr::select(.data$row_number, .data$horizon, groups, names(lag_functions)) %>%
+                dplyr::as_tibble()
 
             } else {  # Assign NA values to all future dynamic features. There's some unnecessary computation here.
 
+              #data_dt <- dtplyr::lazy_dt(data[, var_names[j], drop = FALSE])
+              #data_dt <- data[, var_names[j], drop = FALSE]
+
               data_x <- data[, var_names[j], drop = FALSE] %>%
                 dplyr::mutate("row_number" = 1:dplyr::n()) %>%
                 dplyr::mutate("max_row_number" = max(.data$row_number, na.rm = TRUE),
@@ -614,7 +652,7 @@ create_lagged_df <- function(data, type = c("train", "forecast"), outcome_col =
         }
 
         data_x
-      })  # End loop 'j', the creation of lagged features for a given forecast model horizon.
+      }, future.packages = "dplyr", future.seed = 1)  # End loop 'j', the creation of lagged features for a given forecast model horizon.
       #------------------------------------------------------------------------
 
       # Merge all feature-level lags into 1 data.frame.
@@ -677,14 +715,30 @@ create_lagged_df <- function(data, type = c("train", "forecast"), outcome_col =
   attr(data_out, "static_features") <- static_features
 
   if (is.null(groups)) {
-    attr(data_out, "row_indices") <- row_names[-(1:lookback_max)]
-    if (is.null(dates)) {
-      attr(data_out, "data_start") <- lookback_max + 1  # Removes NAs at the beginning of the dataset
-      attr(data_out, "data_stop") <- max(row_names, na.rm = TRUE)
+    if (isFALSE(keep_rows)) {
+      attr(data_out, "row_indices") <- row_names[-(1:lookback_max)]
     } else {
-      attr(data_out, "date_indices") <- dates
+      attr(data_out, "row_indices") <- row_names
+    }
+    if (is.null(dates)) {
+      if (isFALSE(keep_rows)) {
+        attr(data_out, "data_start") <- lookback_max + 1  # Removes NAs at the beginning of the dataset
+      } else {
+        attr(data_out, "data_start") <- min(row_names, na.rm = TRUE)  # Keep NAs at the beginning of the dataset
+      }
+      attr(data_out, "data_stop") <- max(row_names, na.rm = TRUE)
+    } else {  # Non-grouped time series with dates.
+      if (isFALSE(keep_rows)) {
+        attr(data_out, "date_indices") <- dates[-(1:lookback_max)]
+      } else {
+        attr(data_out, "date_indices") <- dates
+      }
       attr(data_out, "frequency") <- frequency
-      attr(data_out, "data_start") <- min(dates[lookback_max + 1], na.rm = TRUE)  # Removes NAs at the beginning of the dataset
+      if (isFALSE(keep_rows)) {
+        attr(data_out, "data_start") <- min(dates[lookback_max + 1], na.rm = TRUE)  # Removes NAs at the beginning of the dataset
+      } else {
+        attr(data_out, "data_start") <- min(dates, na.rm = TRUE)  # Keep NAs at the beginning of the dataset
+      }
       attr(data_out, "data_stop") <- max(dates, na.rm = TRUE)
     }
   } else {  # Grouped data requires a 'dates' argument.
@@ -746,13 +800,14 @@ summary.lagged_df <- function(object, ...) {
 #' a list of plots, one per feature, of class 'ggplot' if \code{lookback_control} was specified.
 #' @example /R/examples/example_plot_lagged_df.R
 #' @export
-plot.lagged_df <- function(x, ...) {
+plot.lagged_df <- function(x, ...) { # nocov start
 
   if (!methods::is(x, "lagged_df")) {
     stop("This method takes an object of class 'lagged_df' as input. Run create_lagged_df() first.")
   }
 
   data <- x
+  rm(x)
 
   horizons <- attributes(data)$horizons
 
@@ -885,4 +940,4 @@ plot.lagged_df <- function(x, ...) {
         labs(fill = NULL) + ggtitle(paste0("Map of Feature Lags: ", predictor_names[i]))
     })
   }
-}
+} # nocov end
